/*
 *  WIN2.C
 *  windowing functions with six bugs fixed
 *  and debugging functions and _L_ macro removed
 */

#include <STDIO.H>
#include <CONIO.H>
#include <STRING.H>
#include <STDLIB.H>
#include <L_MEM.H>

typedef struct wndw_t
    {
    int w_ur;              /*  upper row of window   */
    int w_lc;              /*  left col of window    */
    int w_num_rows;        /*  no. of window rows    */
    int w_num_cols;        /*  no. of window cols    */
    int frame_width;       /*  width of frame        */
    int t_attr;            /*  text attribute        */
    int fr_attr;           /*  frame attribute       */
    /*  values above are defined, below calculated   */
    int w_lr;              /*  lower row of window   */
    int w_rc;              /*  right col of window   */
    int t_num_rows;        /*  number of text rows   */
    int t_num_cols;        /*  number of text cols   */
    int far * far *backgr; /*  ptr to background     */
    int far * far *buffer; /*  ptr to window buffer  */
    int far * far *text;   /*  ptr to text area      */
    char handle;           /*  handle for window     */
    };

#define NUM_WINDOWS 4
#define HIDE_CURSOR TRUE
#define COLOR       TRUE
/*  */

#if COLOR
#define SCREEN 0xB8000000
#else
#define SCREEN 0xB0000000
#endif

struct wndw_t wndw[NUM_WINDOWS] =
    {
        {
        #if HIDE_CURSOR         /*  original screen  */
        0, 0, 25, 80, 0, 0, 0   /*  window 0  */
        /*  text attribute is black on black  */
        #else
        0, 0, 25, 80, 0, 7, 0
        /*  text attribute is white on black  */
        #endif
        },
        {
        5, 5, 10, 64, 2, 43, 112    /*  window 1  */
        },
        {
        10, 14, 8, 33, 1, 31, 79    /*  window 2  */
        },
        {
        6, 23, 13, 26, 0, 94, 0     /*  window 3  */
        }
    };

int err_flag;                 /*  global error flag  */
char far * far *screen_state; /*  windows visible    */
int far * far *screen;        /*  screen as array    */
char msg[81];

void main(void);
int open_window(struct wndw_t far *w);
void w_clear(struct wndw_t far *w);
void refresh_window_display(struct wndw_t far *w);
int close_window(struct wndw_t far *w);
void w_print(struct wndw_t far *w, int row,
    int col, char *str);
void press_a_key(int i);

/*-----------*/
void main(void)
{
int i,r,term_errors=ALL_ERRORS;
char minus_one=-1;

#if !COLOR
/*  adjust attributes for monochrome monitor  */
for ( i=0; i<NUM_WINDOWS; i++ )
    {
    wndw[i].t_attr = 7;
    wndw[i].fr_attr = 120;
    }
#endif

set_termination_errors(&term_errors);
/*
 *
 */

/*  set up screen state buffer and initialize  */
screen_state =
    (char far * far *)array_alloc(&err_flag,
        sizeof(char),&minus_one,2,25,80);

/*  set up simulated array for screen  */
screen = (int far * far *)array2_sim_alloc(
  (void far *)SCREEN,25,80,sizeof(int),NULL,&err_flag);

open_window(&wndw[0]);
/*  saves the original screen; now open the windows  */

for ( i=1; i<NUM_WINDOWS; i++ )
    {
    open_window(&wndw[i]);
    /*  write strings to the window buffer  */
    sprintf(msg,"Window %d",i);
    for ( r=0;
        r<wndw[i].t_num_rows-!wndw[i].frame_width; r++)
        w_print(&wndw[i],r,2*r,msg);
    refresh_window_display(&wndw[i]);
    press_a_key(i);
    if ( i > 1 )
        {
        /*  write to preceding window  */
        w_print(&wndw[i-1],3,0,"This was written while"
            " window partially covered.");
        refresh_window_display(&wndw[i-1]);
        press_a_key(i);
        }
    }

/*  close the windows  */
for ( i=NUM_WINDOWS-1; i>=0; i-- )
    {
    close_window(&wndw[i]);
    if ( i > 1 )
        press_a_key(i-1);
    }

array_free(screen_state,&err_flag);
array_free(screen,&err_flag);
}

/*---------------------------------*/
int open_window(struct wndw_t far *w)
{
int r,c, char_attr;
static char last_wndw_handle = -1;

/*  first save background  */
w->backgr = (int far * far *)array_alloc(&err_flag,
    sizeof(int),NULL,2,w->w_num_rows,w->w_num_cols);

if ( err_flag )
    return ( err_flag );

w->w_lr = w->w_ur + w->w_num_rows - 1;
/*  save background  */
for ( r=w->w_ur; r<=w->w_lr; r++ )
    {
    /*
     */
    memcpy_f(w->backgr[r-(w->w_ur)],&screen[r][w->w_lc],
        w->w_num_cols*sizeof(int),&err_flag);
    /*
     */
    }

/*  allocate space for window buffer and initialize  */
char_attr = (w->fr_attr<<8) + ' ';
w->buffer = (int far * far *)array_alloc(&err_flag,
 sizeof(int),&char_attr,2,w->w_num_rows,w->w_num_cols);

if ( err_flag )
    return ( err_flag );

/* */
w->w_rc = w->w_lc + w->w_num_cols - 1;
w->t_num_rows = w->w_num_rows - 2*w->frame_width;
w->t_num_cols = w->w_num_cols - 4*w->frame_width;

/*  represent text area as 2-d array  */
w->text =
    (int far * far *)array2_sim_alloc(w->buffer[0]
        + w->frame_width*w->w_num_cols
        + w->frame_width*2,w->t_num_rows,w->w_num_cols,
        sizeof(int),NULL,&err_flag);

w_clear(w);
/*  give this window a handle  */
w->handle = ++last_wndw_handle;

/*  update screen state buffer  */
for ( r=w->w_ur; r<=w->w_lr; r++ )
    for ( c=w->w_lc; c<=w->w_rc; c++ )
        screen_state[r][c] = w->handle;

/*  copy new window to screen  */
refresh_window_display(w);
return ( 0 );
}

/*  clears text area of window  */
/*-------------------------------*/
void w_clear(struct wndw_t far *w)
{
int r;

/*  clear first line of text area of window buffer  */
memset_int_f(w->text[0],(w->t_attr<<8)|0x20,
    w->t_num_cols,&err_flag);

if ( !err_flag )
    for ( r=1; r<w->t_num_rows; r++ )
        memcpy_f(w->text[r],w->text[0],
            2*w->t_num_cols,&err_flag);
}

/*---------------------------------------------*/
void refresh_window_display(struct wndw_t far *w)
{
register int r,c;
/*  */

/*  copy to screen those parts of the buffer
 *  which should be visible
 */
for ( r=w->w_ur; r<=w->w_lr; r++ )
    {
    for ( c=w->w_lc; c<=w->w_rc; c++ )
        {
        if ( screen_state[r][c] == -1 )
            {
            screen_state[r][c] = w->handle;
            /* */
            }
        if ( screen_state[r][c] == w->handle )
            {
            screen[r][c] = w->buffer[r-(w->w_ur)][c-(w->w_lc)];
            /*
             */
            }
        }
    }
}

/*----------------------------------*/
int close_window(struct wndw_t far *w)
{
register int r,c;
int i;

/*  copy background to screen  */
    for ( r=w->w_ur; r<=w->w_lr; r++ )
        {
        /*  */
        memcpy_f(&screen[r][w->w_lc],w->backgr[r-(w->w_ur)],
            w->w_num_cols*sizeof(int),&err_flag);
        }

for ( r=w->w_ur; r<=w->w_lr; r++ )
    for ( c=w->w_lc; c<=w->w_rc; c++ )
        screen_state[r][c] = -1;

for ( i=w->handle-1; i>0; i-- )
    refresh_window_display(&wndw[i]);

array_free(w->text,&err_flag);
if ( !err_flag )
    {
    array_free(w->buffer,&err_flag);
    if ( !err_flag )
        array_free(w->backgr,&err_flag);
    }

return ( err_flag );
}

/*------------------------------*/
void w_print(struct wndw_t far *w,
             int row,
             int col,
             char *str)
{
int attr,j,s=strlen(str);

/*  determine number of bytes to copy
 *  so as not to write outside of text area
 */
if ( col + s > w->t_num_cols )
	s = w->t_num_cols - col;

if ( s > 0 )
    {
    for ( j=0; j<s; j++ )
        {
        attr = w->text[row][col+j] & 0xFF00;
        w->text[row][col+j] = attr | str[j];
        }
    }
}

/*-------------------*/
void press_a_key(int i)
{
w_print(&wndw[i],
    wndw[i].t_num_rows-1,0,"Press a key ...     ");
refresh_window_display(&wndw[i]);
getch();
w_print(&wndw[i],
    wndw[i].t_num_rows-1,0,"                    ");
refresh_window_display(&wndw[i]);
}
